import click
import json
from otto.validate import InputValidation
from .utilities import setup_twilio_client, get_attribute_config_items, echo_format_msg, get_assistant_names
from .resources import Assistant, FieldType, Task, ModelBuild


@click.group()
def handler():
    pass


@handler.command()
def init():
    """
    Creates an empty deployment file to serve as a template for defining your chatbot.
    :return:
    """
    copilot_settings = {
        "assistant": {
            "unique_name": "",
            "friendly_name": "",
            "defaults": {
                "defaults": {
                    "assistant_initiation": "",
                    "fallback": "",
                    "collect": {
                        "validate_on_failure": ""
                    }
                }
            }
        },
        "field_type__custom1": {
            "unique_name": "",
            "friendly_name": "",
            "values": []
        },
        "task__custom1": {
            "unique_name": "",
            "friendly_name": "",
            "actions": {
                "actions": [

                ]
            },
            "task_fields": [
                {
                    "unique_name": "",
                    "field_type": ""
                }
            ],
            "samples": []
        },
        "model": {
            "unique_name": ""
        }
    }

    with open("model-deploy-no-tasks.json", "w") as f:
        json.dump(copilot_settings, f, indent='\t')


@handler.command()
@click.argument("config_loc")
@click.option('--overwrite', default=False, is_flag=True)
def deploy(config_loc, overwrite):
    """
    Deploys a Twilio Autopilot model based on the configuration file provided.

    :param config_loc: Location of the configuration file to deploy.
    :param overwrite: Boolean flag indicating if a model should be overwritten if it exists.
    :return:
    """
    # Setup the Twilio client with the provided authorization.
    response = setup_twilio_client()
    if not response["STATUS"]:
        echo_format_msg(response["Message"])
        exit()

    client = response["Payload"]

    # Load and validate the configuration file.
    config = json.load(open(config_loc, "r"))
    input_validation = InputValidation(config)

    if not input_validation.validate_input():
        click.echo(click.style("DEPLOY FAILED.  See above for areas to improve.", fg='red'), nl=True)
        exit()
    else:
        click.echo("\n")
        click.echo(click.style("VALIDATION PASSED!", fg="green"), nl=True)

    # Get the assistant if it exists or create a new one.
    assistant = Assistant(client.autopilot, **config["assistant"])

    if (assistant.exists() is True) & (overwrite is False):
        response = input("Assistant already exists."
                         "Overwriting the assistant will create a new assistant based on your configuration, which"
                         "could cause some existing resources to be deleted.\n\n"
                         "Do you want to overwrite the existing assistant (y/n)? ")

        if response.lower() != "y":
            msg = "DEPLOY FAILED: Assistant already exists.  You can overwrite the current assistant or create a new" \
                  "one with a different name."
            click.echo(click.style(msg, fg="red"))
            exit()

    assistant = assistant.fetch_or_create()

    # Setup any custom fields that are used by the assistant.
    for field_type_key, field_type_definition in get_attribute_config_items(config, "field_types__").items():
        field_type = FieldType(**field_type_definition).create(assistant)
        msg = "COMPLETED: Custom field type {} has been created.".format(field_type.unique_name)
        echo_format_msg(msg)

    # Add Tasks to the chatbot.
    config_tasks = {k: v for (k, v) in config.items() if k.startswith("task__")}
    for task_key, task_definition in config_tasks.items():
        task = Task(**task_definition).create(assistant)
        msg = "COMPLETED: Task {} has been created.".format(task.unique_name)
        echo_format_msg(msg)

    # Now train the model.
    model = ModelBuild(**config.get("model")).create(assistant)
    msg = "COMPLETED: Model {} has been created.".format(model.unique_name)
    echo_format_msg(msg)

    msg = "SUCCESS!  Your Twilio Autopilot Assistant '{}' has been deployed!".format(assistant.unique_name)
    click.echo(click.style(msg, fg="green"))


@handler.command()
@click.argument("autopilot_sid")
def teardown(autopilot_sid):
    """
    Tears down an Autopilot bot working through the resource hierarchy.  All resources associated with the bot
    will be deleted.

    :param autopilot_sid: The unique identifier of the bot.
    :return:
    """
    client = setup_twilio_client()

    # Get the assistant.  Raise error if doesn't exist.
    try:
        assert autopilot_sid in get_assistant_names(client)
    except AssertionError:
        msg = "FAIL: There is no assistant with the identifier {} to teardown.  Check if the Assistant exists and " \
              "try again.".format(autopilot_sid)
        echo_format_msg(msg)
        exit()

    assistant = client.autopilot.assistants(autopilot_sid).fetch()

    # Begin by removing tasks.
    for t in assistant.tasks.list():

        # Remove all samples.
        for s in t.samples.list():
            s.delete()

        # Remove all fields.
        for f in t.fields.list():
            f.delete()
        t.delete()

    # Remove any custom field types.
    for field_type in assistant.field_types.list():

        for field_value in field_type.field_values.list():
            field_value.delete()

        field_type.delete()

    # Remove models.
    for model in assistant.model_builds.list():
        model.delete()

    assistant.delete()

    msg = "SUCCESS! The '{}' assistant and all of its related resources have been deleted.".format(autopilot_sid)
    click.echo(click.style(msg, fg="green"))


if __name__ == "__main__":
    handler()



import os
import click
from twilio.rest import Client
from twilio.base.exceptions import TwilioException


def get_attribute_config_items(config, attribute):
    return {k: v for (k, v) in config.items() if k.startswith(attribute)}


def setup_twilio_client():
    """
    Sets up the Twilio client that will be used based on the authentication of the user.
    :return: Returns a Twilio Client object.
    """
    response = {
        "STATUS": "PASS"
    }
    try:
        account_sid = os.environ["TWILIO_ACCOUNT_SID"]
        auth_token = os.environ["TWILIO_AUTH_TOKEN"]

        client = Client(account_sid, auth_token)

        # Capture if authentication failed.
        client.autopilot.assistants.list()

        response["Payload"] = client
    except KeyError as e:
        response["STATUS"] = "FAIL"
        response["Message"] = "FAIL: Check that your Twilio environmental variables have been configred correctly."
    except TwilioException:
        response["STATUS"] = "FAIL"
        response["Message"] = "FAIL: Double check your Twilio credentials are correct."

    return response


def get_assistant_names(client):
    """
    Gets the names of all the assistants associated with the Twilio client.
    :param client: Twilio client object.
    """
    return [assistant.unique_name for assistant in client.autopilot.assistants.list()]


def echo_format_msg(msg):
    """
    Colors a message according to the status that is implied.
    :param msg: Text of message to color.
    """
    color_map = {
        "INFO": "blue",
        "FAIL": "red",
        "COMPLETED": "green",
        "SUCCESS": "green"
    }

    intent = msg.split(":")[0]

    click.echo(click.style(msg, fg=color_map[intent]), nl=True)



import re
import click
from collections import Counter
from otto.utilities import echo_format_msg


class InputValidation:

    def __init__(self, config):
        self.config = config
        self.validation_pass = True

    def required_parameters(self):
        """
        Each input file must have:
            - An assistant resource.
            - At least 1 task resource.
            - A Model Build resource.
        """
        click.echo("CHECKING BASE FILE FORMAT....")
        required_tags = ["assistant", "model"]
        missing_tags = list(set(required_tags) - set(self.config.keys()))

        if missing_tags:
            missing_str = ", ".join(missing_tags)
            msg = "FAIL: Configuration file is missing the required element(s): {}".format(missing_str)
            echo_format_msg(msg)
            self.validation_pass = False

        if not any(k.startswith("task__") for k in self.config.keys()):
            msg = "FAIL: Your Assistant has no tasks defined.  Must have at least 1 task."
            echo_format_msg(msg)
            self.validation_pass = False

    def validate_task(self, lbl, task):
        """
        Each task must have:
            - At least 1 sample.
            - A Field if custom field identified in samples
            - If a custom field exists in a sample there must be a corresponding field resource in the task
            - If task has a `collect` action make sure `on_complete` is a `redirect` action.
            - A task cannot have multiple instances of the same action.
        """
        if "actions" not in task:
            msg = "INFO: Task `{}` has no actions associated with it.".format(lbl)
            echo_format_msg(msg)
        elif "actions" not in task["actions"]:
            msg = "FAIL: Actions object in task `{}` is misformed.".format(lbl)
            echo_format_msg(msg)
            self.validation_pass = False
        else:
            task_actions = task["actions"]["actions"]
            for ta in task_actions:
                try:
                    collect_action = ta["collect"]["on_complete"]
                    assert list(collect_action.keys()) == ["redirect"]
                except KeyError:
                    continue
                except AssertionError:
                    collect_action = ta["collect"]["on_complete"]
                    str_exists = ", ".join(collect_action.keys())
                    msg = "FAIL: In `{}` the `on_complete` action for `collect` must be a `redirect`.  You used a `{}` action".\
                        format(lbl, str_exists)
                    echo_format_msg(msg)
                    self.validation_pass = False

            # Check if actions are repeated in the task.
            action_counts = Counter([list(action.keys())[0] for action in task_actions])
            for action, counts in action_counts.items():
                if counts > 1:
                    msg = "FAIL: In task `{}` you have multiple `{}` actions. Can only have one.".format(lbl, action)
                    echo_format_msg(msg)
                    self.validation_pass = False

        # Check if task has samples associated with it.
        try:
            n_samples = len(task["samples"])

            if n_samples == 0:
                msg = "FAIL: There are no samples provided for '{}'".format(lbl)
                echo_format_msg(msg)
                self.validation_pass = False
            elif n_samples < 10:
                msg = "INFO: Task `{}` only has `{}` samples.  Recommended to have at least 10.".format(lbl, str(n_samples))
                echo_format_msg(msg)
        except KeyError:
            msg = "FAIL: There are no samples provided for '{}'".format(lbl)
            echo_format_msg(msg)
            self.validation_pass = False

        # Check that task fields are defined.
        for task_field in task.get("task_fields", []):
            if "unique_name" not in task_field:
                msg = "FAIL: A task field in '{}' is missing a unique_name.".format(lbl)
                echo_format_msg(msg)
                self.validation_pass = False

            if "field_type" not in task_field:
                msg = "FAIL: A task field in '{}' is missing a field_type.".format(lbl)
                echo_format_msg(msg)
                self.validation_pass = False

        # Make sure samples with custom fields have a defined field.
        missing_fields = []
        task_field_labels = [tf["unique_name"] for tf in task.get("task_fields", [])]
        for sample in task.get("samples", []):
            if isinstance(sample, dict):
                sample = sample["tagged_text"]

            for custom_field in re.findall(r"\{(\w+)\}", sample):
                if custom_field not in task_field_labels:
                    missing_fields.append(custom_field)

        if missing_fields:
            missing_fields_str = ", ".join(list(set(missing_fields)))
            msg = "FAIL: Task `{}` contains undefined custom fields.  Make sure your task includes these fields: `{}`".\
                format(lbl, missing_fields_str)
            echo_format_msg(msg)
            self.validation_pass = False

    def validate_assistant(self, lbl, assistant):
        """
        Check the assistant for properly formatted input.

        Specifically checks how the `defaults` parameters is formatted if it is used.
        """
        # Check that defaults are defined correctly in Assistant.
        if "defaults" in assistant:
            if "defaults" not in assistant["defaults"]:
                msg = "FAIL: The `defaults` parameter for Assistant `{}` is not formed properly.".format(lbl)
                echo_format_msg(msg)
                self.validation_pass = False
            else:
                if not all(x in assistant["defaults"]["defaults"] for x in ["assistant_initiation", "fallback"]):
                    msg = "INFO: For Assistant `defaults` to take effect you must minimally specify a " \
                          "default for `assistant_initiation` and `fallback`.  Check Assistant `{}` " \
                          "for missing parameters".format(lbl)
                    echo_format_msg(msg)

    def validate_field_type(self, lbl, assistant):
        """
        Check that custom field types have values associated with them.  Will not cause a critical error, but does
        alert when values are missing since they aren't really useful without defining values.
        """
        # Raise a warning if no values associated with a field type.
        if ("values" not in assistant) or (len(assistant["values"]) == 0):
            msg = "INFO: Custom Field Type `{}` has no values associated with it.".format(lbl)
            echo_format_msg(msg)

    def validate_input(self):
        """
        Performs the validation on the input configuration file.
        :return: Boolean indicating if a critical error is detected in the input file.
        """
        self.required_parameters()

        # Iterate across tags to check tag specific criteria.
        click.echo("\nCHECKING INDIVIDUAL RESOURCES...")
        for tag, attributes in self.config.items():
            click.echo("\nValidating `{}`...".format(tag))
            if "unique_name" not in attributes:
                msg = "FAIL: `{}` does not have a 'unique_name'".format(tag)
                echo_format_msg(msg)
                self.validation_pass = False

            if tag.startswith("task__"):
                self.validate_task(tag, attributes)
            elif tag == "assistant":
                self.validate_assistant(tag, attributes)
            elif tag.startswith("field_type__"):
                self.validate_field_type(tag, attributes)

        return self.validation_pass



from setuptools import setup


with open('README.md') as readme_file:
    long_description = readme_file.read()


setup(
    name="otto-bot",
    version="0.0.1",
    packages=['otto'],
    install_requires=["twilio==6.26.1", "click==7"],
    python_requires='>=3.6',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/mvielkind/otto-bot",
    entry_points={
        'console_scripts': [
            'otto-bot=otto.cli:handler'
        ]
    },
    author="Matthew Vielkind",
    author_email="matt@righthandcode.com",
    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        "License :: OSI Approved :: MIT License",
    ]
)



import re
import logging
from twilio.base.exceptions import TwilioRestException
from otto.utilities import get_attribute_config_items


def get_resource_params(resource_obj, params):
    return {k: v for (k, v) in resource_obj.items() if k in params}


def teardown_nested_resources(base_resource, nested_resources):
    for resource_type in nested_resources:
        for resource_obj in getattr(base_resource, resource_type).list():
            resource_obj.delete()


class Assistant:

    RESOURCE_PARAMS = ["unique_name", "friendly_name", "log_queries", "style_sheet", "defaults"]

    def __init__(self, twilio_client, unique_name, friendly_name=None, log_queries=True, style_sheet=None,
                 defaults=None):
        """
        An object representing an Autopilot assistant.  Modeled after the Assistant resource from Twilio.

        Twilio Assistant Documentation: https://www.twilio.com/docs/autopilot/api/assistant

        :param twilio_client: The Twilio Client object that will create the assistant.
        :param unique_name: str.  Name given to uniquely identify the assistant.
        :param friendly_name: str.  A user friendly name to assign to the assistant.
        :param log_queries: bool.  Indicates whether queries should be logged or not.
        :param style_sheet: str.  A JSON string identifying the Assistant's stylesheet.
        :param defaults: dict.  A JSON  object defining the default tasks for the Assistant.
        """
        self.twilio_client = twilio_client
        self.unique_name = unique_name
        self.friendly_name = friendly_name
        self.log_queries = log_queries
        self.style_sheet = style_sheet
        self.defaults = defaults

    def validate(self):
        response = {
            "FAIL": [],
            "WARN": []
        }
        # Check that `defaults` attribute is constructed correctly.
        if self.defaults is None:
            return response
        else:
            if "defaults" not in self.defaults:
                response["FAIL"].append("ERROR: The `defaults` parameter for Assistant {} is not formed properly.").\
                    format(self.unique_name)
            else:
                if not all(x in self.defaults["defaults"] for x in ["assistant_initiation", "fallback"]):
                    response["WARN"].append("INFO: For Assistant `defaults` to take effect you must minimally specify a"
                                            "default for `assistant_initiation` and `fallback`.  Check Assistant `{}` "
                                            "for missing parameters").format(self.unique_name)

        return response

    def exists(self):
        """
        Checks whether an assistant exists.
        :return: Boolean indicating if the object exists or not.
        """
        try:
            self.twilio_client.assistants(self.unique_name).fetch()
            return True
        except TwilioRestException:
            return False

    def fetch_or_create(self):
        """
        If an assistant exists then update the assistant with the current parameters and returns the Assistant.

        If the assistant does not exist then create it based on the current values.
        :return: A Twilio AssistantInstance object.
        """
        assistant_params = get_resource_params(self.__dict__, self.RESOURCE_PARAMS)

        try:
            assistant = self.twilio_client.assistants(self.unique_name).fetch()

            # Remove any existing tasks and fields from current assistant.
            for task in assistant.tasks.list():
                Task(task.unique_name).teardown(assistant)

            for field_type in assistant.field_types.list():
                FieldType(field_type.unique_name).teardown(assistant)

            return self.twilio_client.assistants(assistant.sid).update(**assistant_params)
        except TwilioRestException:
            return self.twilio_client.assistants.create(**assistant_params)


class Task:

    RESOURCE_PARAMS = ["unique_name", "friendly_name", "actions", "actions_url"]
    NESTED_RESOURCES = ["samples", "fields"]

    def __init__(self, unique_name, samples=None, friendly_name=None, actions=None, actions_url=None, task_fields=None):
        """
        An object representing a Task for an assistant.

        Twilio Task Documentation: https://www.twilio.com/docs/autopilot/api/task

        :param unique_name: str.  Unique name assigned to the task.
        :param samples: list or dict.  Samples to train the task against.  If list is provided will assume language is
            'en-US' by default.
        :param friendly_name: str.  Friendly name defining the task.
        :param actions: dict.  Expects "actions" as a key and then a list of actions defining the task.
        :param actions_url: URL where the task can fetch actions.
        :param task_fields: list.  A list of Field objects if there are field attributes defined for the task.
        """
        self.unique_name = unique_name
        self.samples = samples
        self.friendly_name = friendly_name
        self.actions = actions
        self.actions_url = actions_url
        self.task_fields = task_fields

    def create(self, assistant):
        """
        Creates a task resource for the given assistant.
        :param assistant: The assistant object to create the task within.
        """
        task_definition = get_resource_params(self.__dict__, self.RESOURCE_PARAMS)

        try:
            task = assistant.tasks(self.unique_name).fetch()
            task = task.update(**task_definition)
        except TwilioRestException:
            task = assistant.tasks.create(**task_definition)

        # Remove any existing samples and fields.
        teardown_nested_resources(task, self.NESTED_RESOURCES)

        # Add any task fields.
        defined_fields = []
        if self.task_fields is not None:
            for task_field in self.task_fields:
                tf = TaskField(**task_field).create(task)
                if tf.unique_name not in defined_fields:
                    defined_fields.append(tf.unique_name)

        custom_fields = []
        for sample in self.samples:
            if isinstance(sample, dict):
                s = Sample(**sample).create(task)
            else:
                s = Sample(tagged_text=sample).create(task)

            for custom_field in re.findall(r"\{(\w+)\}", s.tagged_text):
                if custom_field not in custom_fields:
                    custom_fields.append(custom_field)

        # Check if samples include field types that are not defined.
        missing_fields = set(custom_fields) - set(defined_fields)
        if missing_fields:
            str_missing = ", ".join(missing_fields)
            warn_msg = "The following fields are used in the samples, but have not definition: {}".format(str_missing)
            logging.warning(warn_msg)

        return task

    def teardown(self, assistant):
        """
        Deletes a Task along with any other nested resources (samples and fields) associated with it.
        :param assistant: The assistant object to delete the task from.
        """
        task = assistant.tasks(self.unique_name).fetch()

        # Remove any existing samples and fields.
        teardown_nested_resources(task, self.NESTED_RESOURCES)

        task.delete()


class Sample:

    def __init__(self, tagged_text, language="en-US", source_channel=None):
        """
        A Sample is associated with a Task and is used for training your assistant.  A sample represents a way that
        a user might express themselves to complete the task.

        Twilio Sample Documentation: https://www.twilio.com/docs/autopilot/api/task-sample

        :param tagged_text: str.  Text of the sample.  Might include field tags.
        :param language: str.  Language the tagged_text is in.
        :param source_channel: str.  Channel the sample is coming from (i.e. voice, sms, chat etc.)
        """
        self.tagged_text = tagged_text
        self.language = language
        self.source_channel = source_channel

    def create(self, task):
        """
        Creates a text sample to include in the training of the task.
        :param task: Task object to associate the sample with.
        """
        return task.samples.create(**self.__dict__)


class ModelBuild:

    def __init__(self, unique_name):
        """
        The object of a trained Twilio Autopilot model.

        Twilio Documentation: https://www.twilio.com/docs/autopilot/api/model-build

        :param unique_name: str.  Unique name to identify the model.
        """
        self.unique_name = unique_name

    def create(self, assistant):
        """
        Creates a new model for the given assistant.  Creating the model will train the model against the resources
        included with the assistant.
        :param assistant: Assistant object to train the model for.
        """
        try:
            model = assistant.model_builds(self.unique_name).fetch()
            model = model.update(**self.__dict__)
        except TwilioRestException:
            model = assistant.model_builds.create(**self.__dict__)

        return model


class TaskField:

    def __init__(self, field_type, unique_name):
        """
        Represents attributes expected to be used in the task samples that tell the model where to look for given
        attributes that might be provided by the user.

        Twilio Documentation: https://www.twilio.com/docs/autopilot/api/task-field

        :param field_type: str.  The type of field for the attribute.  Could be a built-in type or a custom field type.
        :param unique_name: str.  Unique name for the field.
        """
        self.field_type = field_type
        self.unique_name = unique_name

    def create(self, task):
        """
        Creates a Field Resource for the task that tells the model to look for a specific attribute when training.
        :param task: Task object to create the field for.
        """
        return task.fields.create(**self.__dict__)


class FieldType:

    RESOURCE_PARAMS = ["unique_name", "field_type"]
    NESTED_RESOURCES = ["field_values"]

    def __init__(self, unique_name, values=None, friendly_name=None):
        """
        Defines a custom Field Type that can be used to capture custom attributes.

        Twilio Documentation: https://www.twilio.com/docs/autopilot/api/field-type

        :param unique_name: str.  Unique name to assign to the
        :param values:
        :param friendly_name:
        """
        self.unique_name = unique_name
        self.field_values = values
        self.friendly_name = friendly_name

    def create(self, assistant):
        """
        Creates the custom field type for the assistant.
        :param assistant: Assistant object to create the custom field for.
        """
        field_type_definition = get_resource_params(self.__dict__, self.RESOURCE_PARAMS)

        try:
            field_type = assistant.field_types(self.unique_name).fetch()
            field_type = field_type.update(**field_type_definition)
        except TwilioRestException:
            field_type = assistant.field_types.create(**field_type_definition)

        # Remove any existing samples and fields.
        teardown_nested_resources(field_type, self.NESTED_RESOURCES)

        for fv in self.field_values or []:
            if isinstance(fv, dict):
                FieldValue(**fv).create(field_type)
            else:
                FieldValue(value=fv).create(field_type)

        return field_type

    def teardown(self, assistant):
        """
        Deletes the field type along with any other nested attributes for the field type.
        :param assistant: Assistant object to delete the field type from.
        """
        field_type = assistant.field_types(self.unique_name).fetch()

        # Remove any existing samples and fields.
        teardown_nested_resources(field_type, self.NESTED_RESOURCES)

        field_type.delete()


class FieldValue:

    def __init__(self, value, language="en-US", synonym_of=None):
        """
        Values to associate with a custom Field Type

        Twilio Documentation: https://www.twilio.com/docs/autopilot/api/field-value

        :param value: str.
        :param language: str.  Language the tagged_text is in.
        :param synonym_of:
        """
        self.language = language
        self.value = value
        self.synonym_of = synonym_of

    def create(self, base_resource):
        return base_resource.\
            field_values.\
            create(**self.__dict__)
